@@ -10,10 +10,11 @@ use crate::grammar::Production;
 use crate::grammar::Symbol;
 use crate::grammar::START_RULE_NAME;
 use std::collections::HashMap;
+use std::rc::Rc;
 
-/// Utility for creating [grammar rules](GrammarRule).
+/// Imperative utility for creating a [Grammar].
 ///
-/// Please read the [module documentation](crate::grammar) for more information and examples.
+/// Please read the [module documentation](mod@crate::grammar) for more information and examples.
 pub struct GrammarBuilder {
     current_precedence: usize,
     grammar:            Grammar,
@@ -35,10 +36,10 @@ impl GrammarBuilder {
     }
 
     fn rule_to_symbols(&mut self, rule_name: &str, symbols: Vec<Symbol>) {
-        let rule_name = rule_name.to_string();
-        let production = Production { symbols };
+        let rule_name = Rc::new(rule_name.to_string());
+        let production = Rc::new(Production { symbols });
 
-        if self.grammar.rules.is_empty() && rule_name != START_RULE_NAME {
+        if self.grammar.rules.is_empty() && *rule_name != START_RULE_NAME {
             self.rule_to_rules(START_RULE_NAME, &[&rule_name]);
         }
 
@@ -50,7 +51,7 @@ impl GrammarBuilder {
                 self.grammar.rules.insert(
                     rule_name.clone(),
                     GrammarRule {
-                        name:           rule_name,
+                        name:           rule_name.clone(),
                         disambiguation: None,
                         productions:    vec![production],
                     },
@@ -59,7 +60,7 @@ impl GrammarBuilder {
         }
     }
 
-    /// Map a rule with name `name` to zero or more [crate::lexer::LexerRule].
+    /// Map a rule with name `name` to zero or more lexemes.
     pub fn rule_to_lexemes(
         &mut self,
         lexeme_kind: &str,
@@ -76,7 +77,7 @@ impl GrammarBuilder {
         self
     }
 
-    /// Map a rule with name `name` to zero or more [GrammarRule].
+    /// Map a rule with name `name` to zero or more rules.
     pub fn rule_to_rules(
         &mut self,
         rule_name: &str,
@@ -93,7 +94,7 @@ impl GrammarBuilder {
         self
     }
 
-    /// Create a [Disambiguation]
+    /// Create a disambiguation
     /// with the specified `associativity`,
     /// granting the rules with names `rule_names` equal precedence.
     ///
@@ -154,3 +155,43 @@ impl GrammarBuilder {
         self.grammar.clone()
     }
 }
+
+#[doc(hidden)]
+#[macro_export]
+macro_rules! __grammar_helper {
+    ($grammar:ident $rule_name:literal => empty) => {
+        $grammar.rule_to_rules($rule_name, &[]);
+    };
+    ($grammar:ident $rule_name:literal => rules $($rule_names:literal)*) => {
+        $grammar.rule_to_rules($rule_name, &[$($rule_names),*]);
+    };
+    ($grammar:ident $rule_name:literal => rule $($rule_names:literal)*) => {
+        $grammar.rule_to_rules($rule_name, &[$($rule_names),*]);
+    };
+    ($grammar:ident $rule_name:literal => lexemes $($lexeme_kinds:literal)*) => {
+        $grammar.rule_to_lexemes($rule_name, &[$($lexeme_kinds),*]);
+    };
+    ($grammar:ident $rule_name:literal => lexeme $($lexeme_kinds:literal)*) => {
+        $grammar.rule_to_lexemes($rule_name, &[$($lexeme_kinds),*]);
+    };
+    ($grammar:ident $associativity:expr => rules $($rule_names:literal)*) => {
+        $grammar.disambiguate($associativity, &[$($rule_names),*]);
+    };
+    ($grammar:ident $associativity:path => rule $($rule_names:literal)*) => {
+        $grammar.disambiguate($associativity, &[$($rule_names),*]);
+    };
+}
+
+/// Declarative utility for creating a [Grammar].
+///
+/// Please read the [module documentation](mod@crate::grammar) for more information and examples.
+#[macro_export]
+macro_rules! grammar {
+    ($($target:expr => $action:ident $($args:literal)*);* ;) => {{
+        let mut grammar = santiago::grammar::GrammarBuilder::new();
+
+        $(santiago::__grammar_helper!(grammar $target => $action $($args)*));*;
+
+        grammar.finish()
+    }};
+}
