@@ -12,17 +12,18 @@ use crate::parser::ParserColumn;
 use crate::parser::ParserState;
 use crate::parser::Tree;
 use std::collections::HashSet;
+use std::rc::Rc;
 
 fn predict(column: &mut ParserColumn, rule: &GrammarRule) {
     for production in &rule.productions {
-        let state = ParserState {
+        let new_state = ParserState {
             name:         rule.name.clone(),
             production:   production.clone(),
             dot_index:    0,
             start_column: column.index,
             end_column:   usize::MAX,
         };
-        column.add(state);
+        column.add(new_state);
     }
 }
 
@@ -57,7 +58,7 @@ fn complete(
         .enumerate()
         .filter_map(|(index, st)| match st.next_symbol() {
             Some(Symbol::Rule(name)) => {
-                if name == *state_name {
+                if name == **state_name {
                     Some(index)
                 } else {
                     None
@@ -69,14 +70,14 @@ fn complete(
 
     for index in indexes {
         let st = &columns[state_start_column].states[index];
-        let parser_state = ParserState {
+        let new_state = ParserState {
             name:         st.name.clone(),
             production:   st.production.clone(),
             start_column: st.start_column,
             end_column:   usize::MAX,
             dot_index:    st.dot_index + 1,
         };
-        columns[column_index].add(parser_state);
+        columns[column_index].add(new_state);
     }
 }
 
@@ -84,11 +85,11 @@ fn complete(
 pub fn parse(
     grammar: &Grammar,
     lexemes: &[Lexeme],
-) -> Result<Vec<Tree>, String> {
+) -> Result<Vec<Rc<Tree>>, String> {
     let columns: Vec<ParserColumn> = earley(grammar, lexemes);
 
     for state in &columns.last().unwrap().states {
-        if state.name == START_RULE_NAME && state.completed() {
+        if *state.name == START_RULE_NAME && state.completed() {
             return Ok(build(grammar, lexemes, &columns, state));
         }
     }
@@ -96,7 +97,7 @@ pub fn parse(
     Err(String::new())
 }
 
-/// Parse the provided (Lexemes)(Lexeme) with the given [Grammar]
+/// Parse the provided [Lexemes](Lexeme) with the given [Grammar]
 /// and the [Earley algorithm](https://en.wikipedia.org/wiki/Earley_parser).
 pub fn earley(grammar: &Grammar, lexemes: &[Lexeme]) -> Vec<ParserColumn> {
     let mut columns: Vec<ParserColumn> = (0..=lexemes.len())
@@ -119,18 +120,19 @@ pub fn earley(grammar: &Grammar, lexemes: &[Lexeme]) -> Vec<ParserColumn> {
         })
         .collect();
 
-    let name = START_RULE_NAME.to_string();
+    let name = Rc::new(START_RULE_NAME.to_string());
     columns[0].add(ParserState {
-        name:         name.clone(),
-        production:   grammar.rules.get(&name).unwrap().productions[0].clone(),
+        production: grammar.rules.get(&name).unwrap().productions[0].clone(),
+        name,
         start_column: 0,
-        end_column:   usize::MAX,
-        dot_index:    0,
+        end_column: usize::MAX,
+        dot_index: 0,
     });
 
     for column_index in 0..columns.len() {
         let mut state_index = 0;
         let mut state_len = columns[column_index].states.len();
+        let mut predicted_names = HashSet::new();
 
         while state_index < state_len {
             let state = &columns[column_index].states[state_index];
@@ -140,8 +142,11 @@ pub fn earley(grammar: &Grammar, lexemes: &[Lexeme]) -> Vec<ParserColumn> {
             } else {
                 match state.next_symbol().unwrap() {
                     Symbol::Rule(name) => {
-                        let rule = grammar.rules.get(&name).unwrap();
-                        predict(&mut columns[column_index], rule);
+                        if !predicted_names.contains(&name) {
+                            let rule = grammar.rules.get(&name).unwrap();
+                            predict(&mut columns[column_index], rule);
+                            predicted_names.insert(name);
+                        }
                     }
                     Symbol::Lexeme(kind) => {
                         if column_index + 1 < columns.len()
