@@ -1,7 +1,7 @@
 use santiago::lexer::LexerBuilder;
-use santiago::lexer::LexerRule;
+use santiago::lexer::LexerRules;
 
-pub fn lexer() -> Vec<LexerRule> {
+pub fn lexer_rules() -> LexerRules {
     LexerBuilder::new()
         .pattern(&["INITIAL"], "CHAR", ".", |lexer| lexer.take())
         .finish()
